# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

import os.path
import warnings
from pathlib import Path
from typing import Optional, List, Iterable, Dict, Union, Tuple

import framefile

from vtcff._args_subset import ArgsSubset
from vtcff._codec import Codec
from vtcff._codec_audio_copy import AudioCopy
from vtcff._codec_video_copy import VideoCopy
from vtcff._common import Scale
from vtcff._filter_crop import Crop
from vtcff._filter_pad import Pad
from vtcff._filter_swscale import SwscaleFilter
from vtcff._filter_transpose import Transpose, TransposeFilter
from vtcff._filter_zscale import ZscaleFilter, ColorSpaces
from vtcff._time_span import BeginEndDuration


class VideoCodecNotSpecifiedError(Exception):
    pass


class AudioCodecNotSpecifiedError(Exception):
    pass


def arg_i(path_or_pattern: Union[str, Path]) -> List[str]:
    path_or_pattern = str(path_or_pattern)
    if '*' in path_or_pattern:
        # http://ffmpeg.org/ffmpeg.html#Video-and-Audio-file-format-conversion
        return ["-f",
                "image2",
                "-pattern_type", "glob",
                "-i", path_or_pattern]
    else:
        if os.path.isdir(path_or_pattern):
            path_or_pattern = framefile.directory_to_pattern(
                framefile.Format.percent,
                Path(path_or_pattern))
        assert isinstance(path_or_pattern, str)
        return ["-i", path_or_pattern]


def desynonimize(arg: str) -> str:
    # https://superuser.com/q/835048
    if arg in ('-vcodec', '-c:v'):
        return '-codec:v'
    if arg in ('-acodec', '-c:a'):
        return '-codec:a'
    return arg


class CustomArgs:
    def __init__(self):
        self._before_i: ArgsSubset = ArgsSubset()
        self._after_i: ArgsSubset = ArgsSubset()
        self._video: ArgsSubset = ArgsSubset()
        self._audio: ArgsSubset = ArgsSubset()

    @property
    def before_i(self) -> ArgsSubset:
        return self._before_i

    @property
    def after_i(self) -> ArgsSubset:
        return self._after_i

    @property
    def video(self) -> ArgsSubset:
        return self._video

    @property
    def audio(self) -> ArgsSubset:
        return self._audio


class FfmpegCommand:

    def __init__(self, use_zscale: bool = True):

        self._use_zscale = use_zscale

        self.ffmpeg_exe: Union[Path, str] = "ffmpeg"  # todo test
        """The ffmpeg executable file, the first argument of 
        the generated command."""

        self.src_file: Optional[Union[Path, str]] = None
        self.src_gamma: Optional[float] = None
        self.src_fps: Optional[float] = None

        self.dst_file: Optional[Union[Path, str]] = None

        # the following fields will affect the parameters that have quite
        # vague meaning. For example, when encoding, they will show
        #   yuv422p10le(tv, bt709, progressive)
        # instead
        #   yuv422p10le(tv, bt709/unknown/unknown, progressive)
        # Opinions on the web agree that this is "just meta-data",
        # and they never lead to conversion. However, and as metadata
        # they don't necessarily written anywhere as well. We will
        # pass them to ffmpeg in the hope that sometimes it will serve
        # the cause of good.
        self._dst_color_primaries_meta: Optional[str] = None
        self._dst_color_trc_meta: Optional[str] = None
        self._dst_colorspace_meta: Optional[str] = None

        # for vtcff we require video and audio codecs to be explicitly
        # specified (although for ffmpeg the are optional). But the actual
        # following filed can be set to None. Because user may want
        # to specify codecs with .custom.video, for example.
        # We will check the codecs are set at the very last stage: when
        # we get all out arguments as strings.
        self.dst_codec_video: Optional[Codec] = VideoCopy()
        self.dst_codec_audio: Optional[Codec] = AudioCopy()

        self.dst_pixfmt: Optional[str] = None

        # про -color_range
        # https://trac.ffmpeg.org/ticket/443
        self._dst_color_range_meta: Optional[bool] = None

        self._filter_chain: List = list()

        self.dst_time_range = BeginEndDuration()

        # the FfmpegCommand object can generate parameters it understands.
        # But if you want to add some exotic args, this can be done by placing
        # then in fields of the following object
        self.custom = CustomArgs()

        self.debug = False

    @property
    def override_general(self) -> ArgsSubset:
        warnings.warn("Use .custom.after_i", DeprecationWarning)  # 2021-10
        return self.custom.after_i

    @property
    def override_audio(self) -> ArgsSubset:
        warnings.warn("Use .custom.audio", DeprecationWarning)  # 2021-10
        return self.custom.audio

    @property
    def override_video(self) -> ArgsSubset:
        warnings.warn("Use .custom.video", DeprecationWarning)  # 2021-10
        return self.custom.video

    def _find_or_create_filter(self, obj_type):
        vf = self._find_filter(obj_type)
        if vf is not None:
            return vf
        result = obj_type()
        self._filter_chain.append(result)
        return result

    def _find_filter(self, obj_type):
        for vf in self._filter_chain:
            if isinstance(vf, obj_type):
                return vf
        return None

    def _replace_filter(self, obj_type, new_instance):
        for idx, item in enumerate(self._filter_chain):
            if isinstance(item, obj_type):
                self._filter_chain[idx] = new_instance
                return
        self._filter_chain.append(new_instance)

    def _remove_filter(self, obj_type):
        items_removed = 0
        for idx, item in enumerate(self._filter_chain):
            if isinstance(item, obj_type):
                del self._filter_chain[idx]
                items_removed += 1

        assert 0 <= items_removed <= 1

    def _zscale(self) -> ZscaleFilter:
        """Returns `ZscaleCommand` from the filter chain.
        If there is no such command, creates one."""
        if not self._use_zscale:
            raise Exception("_use_zscale is False")
        return self._find_or_create_filter(ZscaleFilter)

    def _swscale(self) -> SwscaleFilter:
        """Returns `ZscaleCommand` from the filter chain.
        If there is no such command, creates one."""
        if self._use_zscale:
            raise Exception("_use_zscale is True")
        return self._find_or_create_filter(SwscaleFilter)

    def _curr_scale_filter(self) -> Union[SwscaleFilter, ZscaleFilter]:
        if self._use_zscale:
            return self._zscale()
        else:
            return self._swscale()

    @property
    def use_zscale(self):
        return self._use_zscale

    @use_zscale.setter
    def use_zscale(self, x: bool):
        if x == self._use_zscale:
            return

        old_scale = self.scale
        old_src_range_full = self.src_range_full
        old_dst_range_full = self.dst_range_full
        old_src_color_space = self.src_color_space
        old_dst_color_space = self.dst_color_space

        if self._use_zscale:
            self._remove_filter(ZscaleFilter)
        else:
            self._remove_filter(SwscaleFilter)

        self._use_zscale = x
        self.scale = old_scale
        self.src_range_full = old_src_range_full
        self.dst_range_full = old_dst_range_full
        self.src_color_space = old_src_color_space
        self.dst_color_space = old_dst_color_space

    @property
    def scale(self) -> Optional[Scale]:
        if self._use_zscale:
            zs: Optional[ZscaleFilter] = self._find_filter(ZscaleFilter)
            if zs is not None:
                return zs.scaling
            return None

        else:
            sw: Optional[SwscaleFilter] = self._find_filter(
                SwscaleFilter)
            if sw is not None and sw.width is not None:
                assert sw.height is not None
                return Scale(sw.width, sw.height, sw.downscale_only)
            return None

    @scale.setter
    def scale(self, s: Optional[Scale]):

        if self._use_zscale:
            zs: ZscaleFilter = self._find_or_create_filter(ZscaleFilter)
            zs.scaling = s
        else:
            sw: SwscaleFilter = self._find_or_create_filter(SwscaleFilter)
            if s is not None:
                sw.width, sw.height = s.width, s.height
                sw.downscale_only = s.downscale_only
            else:
                sw.width = None
                sw.height = None
                sw.downscale_only = False

    @property
    def transpose(self) -> Optional[Transpose]:
        f: Optional[TransposeFilter] = self._find_filter(TransposeFilter)
        if f is None:
            return None
        else:
            return f.kind

    @transpose.setter
    def transpose(self, x: Transpose):
        f: TransposeFilter = self._find_or_create_filter(TransposeFilter)
        f.kind = x

    @property
    def crop(self) -> Optional[Crop]:
        return self._find_filter(Crop)

    @crop.setter
    def crop(self, x: Crop):
        self._replace_filter(Crop, x)

    @property
    def _pad(self) -> Optional[Pad]:
        """Marked private because not tested."""
        # todo test
        return self._find_filter(Pad)

    @_pad.setter
    def _pad(self, x: Pad):
        """Marked private because not tested."""
        # todo test
        self._replace_filter(Pad, x)

    @property
    def src_color_space(self) -> Optional[str]:
        if self._use_zscale:
            # todo test
            return ColorSpaces.zscale_to_ffmpeg(self._zscale().src_matrix)
        else:
            # todo test
            return self._swscale().src_matrix

    @src_color_space.setter
    def src_color_space(self, val: Optional[str]):
        if self._use_zscale:
            self._zscale().src_matrix = ColorSpaces.ffmpeg_to_zscale(val)
        else:
            # todo test
            self._swscale().src_matrix = val

    @property
    def dst_color_space(self) -> Optional[str]:
        if self._use_zscale:
            # todo test
            return ColorSpaces.zscale_to_ffmpeg(self._zscale().dst_matrix)
        else:
            # todo test
            return self._swscale().dst_matrix

    @dst_color_space.setter
    def dst_color_space(self, val: Optional[str]):
        if self._use_zscale:
            self._zscale().dst_matrix = ColorSpaces.ffmpeg_to_zscale(val)
        else:
            self._swscale().dst_matrix = val

        self._dst_color_primaries_meta = val
        self._dst_color_trc_meta = val
        self._dst_colorspace_meta = val

    @property
    def dst_range_full(self) -> Optional[bool]:
        return self._curr_scale_filter().dst_range_full

    @dst_range_full.setter
    def dst_range_full(self, x: Optional[bool]):
        self._curr_scale_filter().dst_range_full = x
        self._dst_color_range_meta = x

    @property
    def src_range_full(self) -> Optional[bool]:
        return self._curr_scale_filter().src_range_full

    @src_range_full.setter
    def src_range_full(self, x: Optional[bool]):
        self._curr_scale_filter().src_range_full = x

    def _iter_known_before_i(self) -> Iterable[Union[str, Tuple[str, str]]]:
        yield str(self.ffmpeg_exe)

        # для файлов EXR стоит указывать что-то вроде '-gamma 2.2', причем
        # еще до аргумента -i
        if self.src_gamma:
            yield "-gamma", str(self.src_gamma)

        if self.src_fps:
            # в случае сборки видео отдельных файлов-кадров, важно
            # указать частоту кадров _перед_ аргументом -i.
            #
            # Примерно так:
            #   ffmpeg -r 30 -f image2 -pattern_type glob -i '/path/*.png'
            #   ffmpeg -r 30 -i '/path/%04d.png'
            #
            # Параметр -r, указанный позже, будет уже касаться более
            # поздних стадий обработки видео. При этом ffmpeg будет
            # читать исходные файлы, предполагая, что каждый из них
            # имеет длительность 1/25 сек (т.е. 25 fps). Результирующий
            # файл может иметь обманчивую частоту кадров 30. Но
            # не будет прямого соответствия между исходными кадрами
            # и сохраненными. Например, из 300 файлов-кадров получится
            # 360-кадровое 12-секундное видео 30 fps. Так было бы при
            # конвертировании 25 fps в 30 fps.

            # TL;DR: -framerate is for image sequences
            #
            # -framerate is an input per-file option. It is meant for input
            # formats which don't have a framerate or PTS defined, image
            # sequences being an example (https://stackoverflow.com/a/51224132)
            yield '-framerate', str(self.src_fps)

            # TL;DR: -r is for video files, it changes the rate
            #
            # -r can be either an input or output option. As an input option,
            # it retimes input frames at that rate. As an output option,
            # it will duplicate or drop frames to achieve the given rate
            # (https://stackoverflow.com/a/51224132)
            yield '-r', str(self.src_fps)

    def _iter_known_all_after_i(self) -> Iterable[
        Union[str, Tuple[str, Optional[str]]]]:
        if self.dst_time_range.begin:
            yield "-ss", str(self.dst_time_range.begin)
        if self.dst_time_range.duration is not None:
            yield "-t", str(self.dst_time_range.duration)
        if self._filter_chain:
            vf_str = ','.join(str(f) for f in self._filter_chain)
            if vf_str:
                yield '-vf', vf_str

        if self.dst_codec_video is not None:
            for pair in self.dst_codec_video.args():
                yield pair

        if self.dst_pixfmt:
            yield '-pix_fmt', self.dst_pixfmt

        # странные параметры, которые определяют "метаданные" результирующего
        # видео. В итоге оно при кодировании выглядит например как
        # (tv, bt709, progressive), а не (unknown/unknown/unknown, progressive)

        if self._dst_colorspace_meta is not None:
            yield '-colorspace', self._dst_colorspace_meta
        if self._dst_color_primaries_meta is not None:
            yield '-color_primaries', self._dst_color_primaries_meta
        if self._dst_color_trc_meta is not None:
            yield '-color_trc', self._dst_color_trc_meta

        if self._dst_color_range_meta is not None:
            # https://trac.ffmpeg.org/ticket/443
            # "0" auto, "1" 16-235, "2" 0-255
            yield '-color_range', '2' if self._dst_color_range_meta else '1'

        yield '-movflags', '+write_colr'
        yield ('-sws_flags',
               'spline+accurate_rnd+full_chroma_int+full_chroma_inp')

        if self.dst_codec_audio is not None:
            for pair in self.dst_codec_audio.args():
                yield pair

        if self.debug:
            yield '-loglevel', 'verbose'

        # последним в списке аргументов должно идти имя целевого файла.
        # Но здесь мы его не возвращаем - метод __iter__ добавит перед именем
        # файла еще что-то - и потом сам допишет имя

    def _combine_overrides(self, src: Iterable[Dict[str, Optional[str]]]) \
            -> Dict[str, Optional[str]]:
        overrides: Dict[str, Optional[str]] = dict()
        for ov in src:
            for k, v in ov.items():
                # todo test overriding overrides
                if k in overrides:
                    print(f"Overriding [{k} {overrides[k]}] with [{k}, {v}]")
                overrides[k] = v
        return overrides

    def _iter_replacing_overrides(
            self,
            args_or_pairs: Iterable[Union[str, Tuple[str, Optional[str]]]],
            overrides: Dict[str, Optional[str]]) \
            -> Iterable[str]:

        returned_keys = set()

        k: str
        v: Optional[str]

        for item in args_or_pairs:
            if isinstance(item, str):
                yield item
            elif isinstance(item, tuple):
                k, v = item
                if k in overrides:
                    other_v = overrides[k]
                    print(f"Overriding [{k} {v}] with [{k}, {other_v}]")
                    yield k
                    if other_v is not None:
                        yield other_v
                else:
                    yield k
                    if v is not None:
                        yield v
                returned_keys.add(k)
            else:
                raise TypeError

        # до сих пор мы генерировали последовательность известных объекту
        # аргументов в привычном нам порядке. Некоторые аргументы могли быть
        # подменены на значения из словаря overrides. Но возможно в словаре
        # overrides есть и другие аргументы, до сих пор не упомянутые.
        # Возвращаем теперь их
        for k, v in overrides.items():
            if k not in returned_keys:
                yield k
                if v is not None:
                    yield v

    def _iter_almost_final(self) -> Iterable[str]:
        """Возвращает аргументы к команде ffmpeg списком."""

        for x in self._iter_replacing_overrides(
                self._iter_known_before_i(),
                dict(self.custom.before_i.pairs())):
            yield x

        if self.src_file:  # todo это не должно быть опциональным
            for arg in arg_i(self.src_file):
                yield arg

        combined_overrides_after_i = self._combine_overrides([
            dict(self.custom.after_i.pairs()),
            dict(self.custom.video.pairs()),
            dict(self.custom.audio.pairs()),
        ])

        for x in self._iter_replacing_overrides(
                self._iter_known_all_after_i(),
                combined_overrides_after_i):
            yield x

        # последним в списке аргументов идет имя целевого файла
        if self.dst_file is None:
            raise ValueError("Output file not specified")
        yield str(self.dst_file)

    def __iter__(self) -> Iterable[str]:
        args = list(desynonimize(arg) for arg in self._iter_almost_final())

        if not any(s in args for s in ("-codec:v", "-vn")):
            raise VideoCodecNotSpecifiedError(args)

        # The -vn / -an / -sn / -dn options can be used to skip inclusion
        # of video, audio, subtitle and data streams respectively
        if not any(s in args for s in ("-codec:a", "-an")):
            raise AudioCodecNotSpecifiedError
        return iter(args)

    def __str__(self):
        return ' '.join(iter(self))



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Optional, Iterable, Tuple

from ._codec import VideoCodec
from ._codec_avc_preset import VcPreset
from ._common import colon_splitted_pairs


class HevcBitrateNotSpecifiedError(Exception):
    pass


class HevcBitrateSpecifiedForLosslessError(Exception):
    pass


class HevcLosslessAndNearLosslessError(Exception):
    pass


class Hevc(VideoCodec):
    def __init__(self,
                 preset: VcPreset = None,
                 lossless: bool = False,
                 near_lossless: bool = False,
                 mbps: float = None):

        self.preset: Optional[VcPreset] = preset
        self.lossless = lossless
        self.near_lossless = near_lossless
        self.mbps = mbps

    def args(self) -> Iterable[Tuple[str, str]]:

        if self.lossless and self.near_lossless:
            raise HevcLosslessAndNearLosslessError

        yield "-codec:v", "libx265"

        params = dict()

        if self.lossless:
            params["lossless"] = "1"
            # we should NOT specify bitrate for lossless encoding
            if self.mbps is not None:
                raise HevcBitrateSpecifiedForLosslessError
        else:
            if self.near_lossless:
                # https://x265.readthedocs.io/en/stable/lossless.html

                params["cu-lossless"] = "1"
                params["psy-rd"] = "1.0"

                # without the following param we may get
                # "x265 [warning]: --cu-lossless disabled,
                # requires --rdlevel 3 or higher"
                params["rd"] = "3"
                # params["ssim"] = "1"  # ?!

            if self.mbps is None:
                raise HevcBitrateNotSpecifiedError
            params["bitrate"] = str(round(self.mbps * 1000))

        p = self.preset
        if p is None:
            if self.lossless:
                # we are not losing any quality here,
                # we want better-than-prores result fast
                p = VcPreset.N1_ULTRAFAST
            elif self.near_lossless:
                # safest bet is highest quality
                p = VcPreset.N10_PLACEBO

        if p is not None:
            yield "-preset", str(p.value)

        if params:
            yield "-x265-params", colon_splitted_pairs(params)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Iterable, Tuple, Optional


class Codec:
    def args(self) -> Iterable[Tuple[str, Optional[str]]]:
        raise NotImplementedError


class VideoCodec(Codec):
    pass


class AudioCodec(Codec):
    pass



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from __future__ import annotations

from enum import Enum, unique
from typing import List


@unique
class VcPreset(Enum):
    # https://trac.ffmpeg.org/wiki/Encode/H.264
    # https://trac.ffmpeg.org/wiki/Encode/H.265

    N1_ULTRAFAST = "ultrafast"
    N2_SUPERFAST = "superfast"
    N3_VERYFAST = "veryfast"
    N4_FASTER = "faster"
    N5_FAST = "fast"
    N6_MEDIUM = "medium"
    N7_SLOW = "slow"
    N8_SLOWER = "slower"
    N9_VERYSLOW = "veryslow"
    N10_PLACEBO = "placebo"

    @classmethod
    def faster(cls, speed: VcPreset) -> VcPreset:
        arr: List[VcPreset] = list(cls)
        idx = arr.index(speed)
        if idx < 0:
            raise ValueError
        if idx == 0:
            raise IndexError
        return arr[idx - 1]



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Optional


class BeginEndDuration:
    def __init__(self):
        self.begin: float = 0
        self.duration: Optional[float] = None

    @property
    def end(self):
        if self.duration is None:
            return None
        else:
            return self.begin + self.duration

    @end.setter
    def end(self, x: Optional[float]):
        if x is None:
            self.duration = None
        else:
            self.duration = x - self.begin



# SPDX-FileCopyrightText: (c) 2016-2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT


from ._command import FfmpegCommand, Scale
from ._filter_pad import Pad


class LetterboxSizes:
    def __init__(self,
                 src_width: int, src_height: int,
                 dst_width: int, dst_height: int):

        src_aspect = src_width / src_height
        dst_aspect = dst_width / dst_height

        needSourceWidth = int(round(dst_aspect * src_height))
        # если бы ширина исходника была равна этому значению,
        # а высота неизменна, то в обоих файлах соотношение сторон было
        # бы одинаковым

        self.pad_left = 0
        self.pad_right = 0
        self.pad_top = 0
        self.pad_bottom = 0

        if needSourceWidth != src_width:
            # придется добавлять полосы
            if src_aspect > dst_aspect:
                # исходник шире, чем результат
                # нужно добавить к исходнику полосы сверху и снизу
                need_source_height = src_width / dst_aspect
                assert need_source_height > src_height
                padding_total = need_source_height - src_height
                padding_half = int(round(padding_total / 2))
                self.pad_top = padding_half
                self.pad_bottom = int(round(padding_total - padding_half))
            else:
                # исходник уже, чем результат
                # нужно добавить к исходнику полосы слева и справа
                assert needSourceWidth > src_width
                padding_total = needSourceWidth - src_width
                padding_half = int(round(padding_total / 2))
                self.pad_left = padding_half
                self.pad_right = padding_total - padding_half


def pad_lrtb_to_lrwh(pad_left=0, pad_right=0, pad_top=0, pad_bottom=0):
    # добавляет черные полоски по бокам видео

    if pad_left < 0:
        raise Exception("padLeft must not be negative.")
    if pad_right < 0:
        raise Exception("padRight must not be negative.")
    if pad_top < 0:
        raise Exception("padTop must not be negative.")
    if pad_bottom < 0:
        raise Exception("padBottom must not be negative.")

    width = "iw"
    height = "ih"
    x = 0
    y = 0

    if any(v > 0 for v in [pad_left, pad_right, pad_top, pad_bottom]):
        width = "iw+%d" % int(pad_left + pad_right)
        height = "ih+%d" % int(pad_top + pad_bottom)
        x = int(pad_left)
        y = int(pad_top)

    return x, y, width, height


def _letterbox(cmd: FfmpegCommand,
               src_width: int, src_height: int,
               dst_width: int, dst_height: int):
    """Marked private because not tested."""
    # todo test
    sizes = LetterboxSizes(src_width=src_width, src_height=src_height,
                           dst_width=dst_width, dst_height=dst_height)
    l, t, w, h = pad_lrtb_to_lrwh(pad_left=sizes.pad_left,
                                  pad_right=sizes.pad_right,
                                  pad_top=sizes.pad_top,
                                  pad_bottom=sizes.pad_bottom)
    cmd._pad = Pad(left=l, top=t, width=w, height=h)
    cmd.scale = Scale(-2, dst_height)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Optional

from vtcff._common import frame_dimension_spec
from vtcff._filter_base import FilterBase
from vtcff._filter_zscale import _full_limited_none, _bool_to_full_limited_none


class SwscaleFilter(FilterBase):
    # it seems, there are at least three ways to convert colors with swscale:
    #   -vf "scale=out_color_matrix..."
    #   -vf "colormatrix=..."
    #   -vf "colorspace=..."
    #
    # https://trac.ffmpeg.org/wiki/colorspace compares "colorspace"
    # and "colormatrix" (2019): "colormatrix produces horrible quality
    # for anything > 8bpc, while colorspace produces something decent,
    # at least for 10bpc. For 8bpc they both produce similar bad quality
    # probably due to improper design in the algorithms. For 8bpc
    # colorspace still seems to produce slightly better quality than
    # colormatrix"
    #
    # However, even if we always preferred `colorspace`, it is not clear how
    # to deal with the color range conversion (full/limited) that is only
    # done with "scale=" (or not?)
    #
    # In general, we choose the simplest way: we use the arguments of the
    # "scale=" filter in the hope that swscale developers will eventually
    # figure out how best to carry out this conversion.
    #
    # But at every opportunity, we prefer not to use this object and
    # swscale at all (zscale seems to be much more predictable)

    def __init__(self):
        super().__init__()
        self.width: Optional[int] = None
        self.height: Optional[int] = None
        self.downscale_only = False

    @property
    def dst_range_full(self) -> Optional[bool]:
        return _full_limited_none(self._pairs.get('out_range'))

    @dst_range_full.setter
    def dst_range_full(self, x: Optional[bool]):
        self._set_or_remove('out_range',
                            _bool_to_full_limited_none(x))

    @property
    def src_range_full(self) -> Optional[bool]:
        return _full_limited_none(self._pairs.get('in_range'))

    @src_range_full.setter
    def src_range_full(self, x: Optional[bool]):
        self._set_or_remove('in_range',
                            _bool_to_full_limited_none(x))

    @property
    def dst_matrix(self) -> Optional[str]:
        # todo test
        return self._pairs.get('out_color_matrix')

    @dst_matrix.setter
    def dst_matrix(self, val: Optional[str]):
        # todo test
        self._set_or_remove('out_color_matrix', val)

    @property
    def src_matrix(self) -> Optional[str]:
        # todo test
        return self._pairs.get('in_color_matrix')

    @src_matrix.setter
    def src_matrix(self, val: Optional[str]):
        # todo test
        self._set_or_remove('in_color_matrix', val)

    def __str__(self) -> str:
        final_pairs = dict(self._pairs)
        if self.width is not None:
            final_pairs["width"] = frame_dimension_spec(
                iw_or_ih='iw',
                value=self.width,
                downscale_only=self.downscale_only)
        if self.height is not None:
            final_pairs["height"] = frame_dimension_spec(
                iw_or_ih='ih',
                value=self.height,
                downscale_only=self.downscale_only)
        # else:
        #     final_pairs["width"] = str(self.width)
        #     final_pairs["height"] = str(self.height)
        return self._to_string("scale", final_pairs.items())



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import NamedTuple


class Crop(NamedTuple):
    left: int = 0
    top: int = 0
    width: int = 0
    height: int = 0

    def __str__(self):
        # def cropRect(self, rectPosX, rectPosY, rectWidth, rectHeight):
        #     cropFilter = "crop=%s:%s:%s:%s" % (
        #         rectWidth, rectHeight, rectPosX, rectPosY)
        return "crop=%s:%s:%s:%s" \
               % (self.width, self.height, self.left, self.top)



import re
from typing import Optional


def _three_digits(text) -> Optional[str]:
    """True if `text` contains something like aXXXb, where XXX are digits
    and a, b - are not."""
    m = re.search(r'(?<!\d)([01234]{3})(?!\d)', text)
    if m:
        return str(m.group(1))
    else:
        return None


def pixfmt_subsampling(pixfmt: str) -> Optional[str]:
    td = _three_digits(pixfmt)
    if td:
        return td
    if any(s in pixfmt for s in ('rgb', 'bgr', 'gbr')):
        return '444'
    return None





# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import NamedTuple, Union


class Pad(NamedTuple):
    # значениями полей может быть размер или что-то вроде "iw+10"
    left: Union[int, str] = 0
    top: Union[int, str] = 0
    width: Union[int, str] = 0
    height: Union[int, str] = 0
    color: str = 'black'

    def __str__(self):
        return "pad=width=%s:height=%s:x=%s:y=%s:color=%s" \
               % (self.width, self.height, self.left, self.top, self.color)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT
import warnings
from typing import Optional

from vtcff._common import Scale, frame_dimension_spec
from vtcff._filter_base import FilterBase


def _bool_to_full_limited_none(x: Optional[bool]):
    if x is None:
        return None
    elif x:
        return "full"
    else:
        return "limited"


def _full_limited_none(x: Optional[str]):
    if x is None:
        return None
    elif x == "full":
        return True
    elif x == "limited":
        return False
    else:
        raise ValueError


class ZscaleFilter(FilterBase):
    def __init__(self):
        super().__init__()
        self.scaling: Optional[Scale] = None

    @property
    def dst_range_full(self) -> Optional[bool]:
        return _full_limited_none(self._pairs.get('range'))

    @dst_range_full.setter
    def dst_range_full(self, x: Optional[bool]):
        self._set_or_remove('range',
                            _bool_to_full_limited_none(x))

    @property
    def src_range_full(self) -> Optional[bool]:
        return _full_limited_none(self._pairs.get('rangein'))

    @src_range_full.setter
    def src_range_full(self, x: Optional[bool]):
        self._set_or_remove('rangein',
                            _bool_to_full_limited_none(x))

    @property
    def dst_matrix(self) -> Optional[str]:
        return self._pairs.get('matrix')

    @dst_matrix.setter
    def dst_matrix(self, val: Optional[str]):
        self._set_or_remove('matrix', val)

    @property
    def src_matrix(self) -> Optional[str]:
        # todo test
        return self._pairs.get('matrixin')

    @src_matrix.setter
    def src_matrix(self, val: Optional[str]):
        # todo test
        self._set_or_remove('matrixin', val)

    # w = -2:h = 2160

    def __str__(self):

        all_pairs = dict()
        if self.scaling is not None:
            all_pairs["filter"] = 'spline36'
            # todo test downscale_only
            all_pairs["w"] = frame_dimension_spec(
                iw_or_ih='iw',
                value=self.scaling.width,
                downscale_only=self.scaling.downscale_only)
            all_pairs["h"] = frame_dimension_spec(
                iw_or_ih='ih',
                value=self.scaling.height,
                downscale_only=self.scaling.downscale_only)

        for k, v in self._pairs.items():
            all_pairs[k] = v

        if not all_pairs:
            return ''

        all_pairs['dither'] = 'error_diffusion'

        return "zscale=" + ":".join(
            lhs + '=' + rhs for (lhs, rhs) in all_pairs.items())


class ColorSpaces:
    # http://underpop.online.fr/f/ffmpeg/help/zscale.htm.gz
    _data = [
        # left ffmpeg, right zscale

        ('bt709', '709'),

        # ffmpeg bt470bg is BT.470BG or BT.601-6 625
        # (http://underpop.online.fr/f/ffmpeg/help/colorspace.htm.gz)
        ('bt470bg', '470bg'),

        # ffmpeg smpte170m is SMPTE-170M or BT.601-6 525
        # (http://underpop.online.fr/f/ffmpeg/help/colorspace.htm.gz)
        ('smpte170m', '170m'),

        # ffmpeg bt2020ncl is BT.2020 with non-constant luminance
        # (http://underpop.online.fr/f/ffmpeg/help/colorspace.htm.gz)
        ('bt2020ncl', '2020_ncl'),

        # todo test ('bt2020' is 8 bit only?)
        # bt2020 - BT.2020
        # bt2020-10 - BT.2020 for 10-bits content
        # bt2020-12 - BT.2020 for 12-bits content
        ('bt2020', '2020_cl', True),
        (None, None)
    ]

    @classmethod
    def _find_by_ffmpeg(cls, x):
        return next(t for t in cls._data if t[0] == x)

    @classmethod
    def _find_by_zscale(cls, x):
        return next(t for t in cls._data if t[1] == x)

    @classmethod
    def _warn_on_need(cls, t):
        if len(t) >= 3 and t[2]:
            warnings.warn(f"colorspaces {t} not tested", stacklevel=3)

    @classmethod
    def ffmpeg_to_zscale(cls, cs):
        try:
            tuple = cls._find_by_ffmpeg(cs)
            cls._warn_on_need(tuple)
            return tuple[1]
        except StopIteration:
            raise ValueError(cs)

    @classmethod
    def zscale_to_ffmpeg(cls, cs):
        try:
            tuple = cls._find_by_zscale(cs)
            cls._warn_on_need(tuple)
            return tuple[0]


        except StopIteration:
            raise ValueError(cs)



from pathlib import Path

from setuptools import setup, find_packages

name = "vtcff"

readme = (Path(__file__).parent / 'README.md').read_text(encoding="utf-8")

setup(
    name=name,
    version='0.3.0',

    author="Artёm IG",
    author_email="ortemeo@gmail.com",
    url='https://github.com/rtmigo/vtcff_py',

    packages=find_packages(include=['vtcff', 'vtcff.*']),

    python_requires='>=3.8',
    install_requires=['framefile'],

    description="Ffmpeg wrapper for transcoding between video formats with "
                "an emphasis on maintaining quality and color depth "
                "in video production pipelines",

    long_description=readme,
    long_description_content_type='text/markdown',

    license="MIT",

    keywords="ffmpeg wrapper "
             "codec encoder decoder transcoding "
             "video audio".split(),

    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Typing :: Typed",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows"
    ],
)